@@ -46,6 +46,15 @@
 #include "utils.h"
 #include "prefs_common.h"
 
+#if USE_THREADS
+G_LOCK_DEFINE_STATIC(mh);
+#define S_LOCK(name)	G_LOCK(name)
+#define S_UNLOCK(name)	G_UNLOCK(name)
+#else
+#define S_LOCK(name)
+#define S_UNLOCK(name)
+#endif
+
 static void	mh_folder_init		(Folder		*folder,
 					 const gchar	*name,
 					 const gchar	*path);
@@ -75,6 +84,15 @@ static gint     mh_add_msgs		(Folder		*folder,
 					 GSList		*file_list,
 					 gboolean	 remove_source,
 					 gint		*first);
+static gint     mh_add_msg_msginfo	(Folder		*folder,
+					 FolderItem	*dest,
+					 MsgInfo	*msginfo,
+					 gboolean	 remove_source);
+static gint     mh_add_msgs_msginfo	(Folder		*folder,
+					 FolderItem	*dest,
+					 GSList		*msglist,
+					 gboolean	 remove_source,
+					 gint		*first);
 static gint     mh_move_msg		(Folder		*folder,
 					 FolderItem	*dest,
 					 MsgInfo	*msginfo);
@@ -151,6 +169,8 @@ static FolderClass mh_class =
 	mh_get_msginfo,
 	mh_add_msg,
 	mh_add_msgs,
+	mh_add_msg_msginfo,
+	mh_add_msgs_msginfo,
 	mh_move_msg,
 	mh_move_msgs,
 	mh_copy_msg,
@@ -208,6 +228,8 @@ static GSList *mh_get_msg_list_full(Folder *folder, FolderItem *item,
 
 	g_return_val_if_fail(item != NULL, NULL);
 
+	S_LOCK(mh);
+
 #ifdef MEASURE_TIME
 	timer = g_timer_new();
 #endif
@@ -290,21 +312,27 @@ static GSList *mh_get_msg_list_full(Folder *folder, FolderItem *item,
 
 		if (newlist == NULL) {
 			procmsg_msg_list_free(mlist);
+			S_UNLOCK(mh);
 			return NULL;
 		}
-		if (mlist == newlist)
+		if (mlist == newlist) {
+			S_UNLOCK(mh);
 			return newlist;
+		}
 		for (cur = mlist; cur != NULL; cur = cur->next) {
 			if (cur->next == newlist) {
 				cur->next = NULL;
 				procmsg_msg_list_free(mlist);
+				S_UNLOCK(mh);
 				return newlist;
 			}
 		}
 		procmsg_msg_list_free(mlist);
+		S_UNLOCK(mh);
 		return NULL;
 	}
 
+	S_UNLOCK(mh);
 	return mlist;
 }
 
@@ -323,6 +351,7 @@ static gchar *mh_fetch_msg(Folder *folder, FolderItem *item, gint num)
 {
 	gchar *path;
 	gchar *file;
+	gchar buf[16];
 
 	g_return_val_if_fail(item != NULL, NULL);
 	g_return_val_if_fail(num > 0, NULL);
@@ -336,7 +365,7 @@ static gchar *mh_fetch_msg(Folder *folder, FolderItem *item, gint num)
 		return NULL;
 
 	path = folder_item_get_path(item);
-	file = g_strconcat(path, G_DIR_SEPARATOR_S, itos(num), NULL);
+	file = g_strconcat(path, G_DIR_SEPARATOR_S, utos_buf(buf, num), NULL);
 	g_free(path);
 	if (!is_file_exist(file)) {
 		g_free(file);
@@ -409,8 +438,8 @@ static gchar *mh_get_new_msg_filename(FolderItem *dest)
 									\
 	if (fp)								\
 		procmsg_write_flags(&newmsginfo, fp);			\
-	else if (dest->opened)						\
-		procmsg_add_flags(dest, n, newmsginfo.flags);		\
+	else 								\
+		procmsg_add_mark_queue(dest, n, newmsginfo.flags);	\
 }
 
 static gint mh_add_msg(Folder *folder, FolderItem *dest, const gchar *file,
@@ -435,8 +464,9 @@ static gint mh_add_msgs(Folder *folder, FolderItem *dest, GSList *file_list,
 	gchar *destfile;
 	GSList *cur;
 	MsgFileInfo *fileinfo;
+	MsgInfo *msginfo;
 	gint first_ = 0;
-	FILE *fp;
+	FILE *fp = NULL;
 
 	g_return_val_if_fail(dest != NULL, -1);
 	g_return_val_if_fail(file_list != NULL, -1);
@@ -446,17 +476,31 @@ static gint mh_add_msgs(Folder *folder, FolderItem *dest, GSList *file_list,
 		if (dest->last_num < 0) return -1;
 	}
 
-	if ((((MsgFileInfo *)file_list->data)->flags == NULL &&
-	    file_list->next == NULL) || dest->opened)
-		fp = NULL;
-	else if ((fp = procmsg_open_mark_file(dest, DATA_APPEND)) == NULL)
-		g_warning("Can't open mark file.\n");
+	S_LOCK(mh);
+
+	if (!dest->opened) {
+		if ((fp = procmsg_open_mark_file(dest, DATA_APPEND)) == NULL)
+			g_warning("mh_add_msgs: can't open mark file.");
+	}
 
 	for (cur = file_list; cur != NULL; cur = cur->next) {
+		MsgFlags flags = {MSG_NEW|MSG_UNREAD, 0};
+
 		fileinfo = (MsgFileInfo *)cur->data;
+		if (fileinfo->flags)
+			flags = *fileinfo->flags;
+		msginfo = procheader_parse_file(fileinfo->file, flags, 0);
+		if (!msginfo) {
+			if (fp) fclose(fp);
+			S_UNLOCK(mh);
+			return -1;
+		}
 
 		destfile = mh_get_new_msg_filename(dest);
-		if (destfile == NULL) return -1;
+		if (destfile == NULL) {
+			S_UNLOCK(mh);
+			return -1;
+		}
 		if (first_ == 0 || first_ > dest->last_num + 1)
 			first_ = dest->last_num + 1;
 
@@ -465,11 +509,14 @@ static gint mh_add_msgs(Folder *folder, FolderItem *dest, GSList *file_list,
 				g_warning(_("can't copy message %s to %s\n"),
 					  fileinfo->file, destfile);
 				g_free(destfile);
+				if (fp) fclose(fp);
+				S_UNLOCK(mh);
 				return -1;
 			}
 		}
 
-		g_signal_emit_by_name(syl_app_get(), "add-msg", dest, destfile, dest->last_num + 1);
+		if (syl_app_get())
+			g_signal_emit_by_name(syl_app_get(), "add-msg", dest, destfile, dest->last_num + 1);
 
 		g_free(destfile);
 		dest->last_num++;
@@ -477,31 +524,25 @@ static gint mh_add_msgs(Folder *folder, FolderItem *dest, GSList *file_list,
 		dest->updated = TRUE;
 		dest->mtime = 0;
 
-		if (fileinfo->flags) {
-			if (MSG_IS_RECEIVED(*fileinfo->flags)) {
-				if (dest->unmarked_num == 0)
-					dest->new = 0;
-				dest->unmarked_num++;
-				procmsg_add_mark_queue(dest, dest->last_num,
-						       *fileinfo->flags);
-			} else {
-				SET_DEST_MSG_FLAGS(fp, dest, dest->last_num,
-						   *fileinfo->flags);
-			}
-			if (MSG_IS_NEW(*fileinfo->flags))
-				dest->new++;
-			if (MSG_IS_UNREAD(*fileinfo->flags))
-				dest->unread++;
-		} else {
+		if (MSG_IS_RECEIVED(flags)) {
+			/* resets new flags of existing messages on
+			   received mode */
 			if (dest->unmarked_num == 0)
 				dest->new = 0;
 			dest->unmarked_num++;
-			dest->new++;
-			dest->unread++;
+			procmsg_add_mark_queue(dest, dest->last_num, flags);
+		} else {
+			SET_DEST_MSG_FLAGS(fp, dest, dest->last_num, flags);
 		}
+		procmsg_add_cache_queue(dest, dest->last_num, msginfo);
+		if (MSG_IS_NEW(flags))
+			dest->new++;
+		if (MSG_IS_UNREAD(flags))
+			dest->unread++;
 	}
 
-	if (fp) fclose(fp);
+	if (fp)
+		fclose(fp);
 
 	if (first)
 		*first = first_;
@@ -514,15 +555,135 @@ static gint mh_add_msgs(Folder *folder, FolderItem *dest, GSList *file_list,
 		}
 	}
 
+	S_UNLOCK(mh);
 	return dest->last_num;
 }
 
+
+static gint mh_add_msg_msginfo(Folder *folder, FolderItem *dest,
+			       MsgInfo *msginfo, gboolean remove_source)
+{
+	GSList msglist;
+
+	g_return_val_if_fail(msginfo != NULL, -1);
+
+	msglist.data = msginfo;
+	msglist.next = NULL;
+
+	return mh_add_msgs_msginfo(folder, dest, &msglist, remove_source, NULL);
+}
+
+static gint mh_add_msgs_msginfo(Folder *folder, FolderItem *dest,
+				GSList *msglist, gboolean remove_source,
+				gint *first)
+{
+	GSList *cur;
+	MsgInfo *msginfo;
+	gchar *srcfile;
+	gchar *destfile;
+	gint first_ = 0;
+	FILE *fp = NULL;
+
+	g_return_val_if_fail(dest != NULL, -1);
+	g_return_val_if_fail(msglist != NULL, -1);
+
+	if (dest->last_num < 0) {
+		mh_scan_folder(folder, dest);
+		if (dest->last_num < 0) return -1;
+	}
+
+	S_LOCK(mh);
+
+	if (!dest->opened) {
+		if ((fp = procmsg_open_mark_file(dest, DATA_APPEND)) == NULL)
+			g_warning("mh_add_msgs_msginfo: can't open mark file.");
+	}
+
+	for (cur = msglist; cur != NULL; cur = cur->next) {
+		msginfo = (MsgInfo *)cur->data;
+
+		destfile = mh_get_new_msg_filename(dest);
+		if (!destfile) {
+			if (fp) fclose(fp);
+			S_UNLOCK(mh);
+			return -1;
+		}
+		if (first_ == 0 || first_ > dest->last_num + 1)
+			first_ = dest->last_num + 1;
+
+		srcfile = procmsg_get_message_file(msginfo);
+		if (!srcfile) {
+			if (fp) fclose(fp);
+			g_free(destfile);
+			S_UNLOCK(mh);
+			return -1;
+		}
+		if (syl_link(srcfile, destfile) < 0) {
+			if (copy_file(srcfile, destfile, TRUE) < 0) {
+				g_warning("mh_add_msgs_msginfo: can't copy message %s to %s", srcfile, destfile);
+				g_free(srcfile);
+				g_free(destfile);
+				if (fp) fclose(fp);
+				S_UNLOCK(mh);
+				return -1;
+			}
+		}
+
+		if (syl_app_get())
+			g_signal_emit_by_name(syl_app_get(), "add-msg", dest, destfile, dest->last_num + 1);
+
+		g_free(srcfile);
+		g_free(destfile);
+		dest->last_num++;
+		dest->total++;
+		dest->updated = TRUE;
+		dest->mtime = 0;
+
+		if (MSG_IS_RECEIVED(msginfo->flags)) {
+			/* resets new flags of existing messages on
+			   received mode */
+			if (dest->unmarked_num == 0)
+				dest->new = 0;
+			dest->unmarked_num++;
+			procmsg_add_mark_queue(dest, dest->last_num,
+					       msginfo->flags);
+		} else {
+			SET_DEST_MSG_FLAGS(fp, dest, dest->last_num,
+					   msginfo->flags);
+		}
+		procmsg_add_cache_queue(dest, dest->last_num, msginfo);
+		if (MSG_IS_NEW(msginfo->flags))
+			dest->new++;
+		if (MSG_IS_UNREAD(msginfo->flags))
+			dest->unread++;
+	}
+
+	if (fp)
+		fclose(fp);
+
+	if (first)
+		*first = first_;
+
+	if (remove_source) {
+		for (cur = msglist; cur != NULL; cur = cur->next) {
+			msginfo = (MsgInfo *)cur->data;
+			srcfile = procmsg_get_message_file(msginfo);
+			if (g_unlink(srcfile) < 0)
+				FILE_OP_ERROR(srcfile, "unlink");
+			g_free(srcfile);
+		}
+	}
+
+	S_UNLOCK(mh);
+	return dest->last_num;
+}
+
+
 static gint mh_do_move_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 {
 	FolderItem *src;
 	gchar *srcfile;
 	gchar *destfile;
-	FILE *fp;
 	GSList *cur;
 	MsgInfo *msginfo;
 
@@ -534,10 +695,7 @@ static gint mh_do_move_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 		if (dest->last_num < 0) return -1;
 	}
 
-	if (dest->opened)
-		fp = NULL;
-	else if ((fp = procmsg_open_mark_file(dest, DATA_APPEND)) == NULL)
-		g_warning(_("Can't open mark file.\n"));
+	S_LOCK(mh);
 
 	for (cur = msglist; cur != NULL; cur = cur->next) {
 		msginfo = (MsgInfo *)cur->data;
@@ -562,8 +720,10 @@ static gint mh_do_move_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 			break;
 		}
 
-		g_signal_emit_by_name(syl_app_get(), "add-msg", dest, destfile, dest->last_num + 1);
-		g_signal_emit_by_name(syl_app_get(), "remove-msg", src, srcfile, msginfo->msgnum);
+		if (syl_app_get()) {
+			g_signal_emit_by_name(syl_app_get(), "add-msg", dest, destfile, dest->last_num + 1);
+			g_signal_emit_by_name(syl_app_get(), "remove-msg", src, srcfile, msginfo->msgnum);
+		}
 
 		g_free(srcfile);
 		g_free(destfile);
@@ -575,10 +735,8 @@ static gint mh_do_move_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 		dest->updated = TRUE;
 		dest->mtime = 0;
 
-		if (fp) {
-			SET_DEST_MSG_FLAGS(fp, dest, dest->last_num,
-					   msginfo->flags);
-		}
+		SET_DEST_MSG_FLAGS(NULL, dest, dest->last_num, msginfo->flags);
+		procmsg_add_cache_queue(dest, dest->last_num, msginfo);
 
 		if (MSG_IS_NEW(msginfo->flags)) {
 			src->new--;
@@ -592,8 +750,12 @@ static gint mh_do_move_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 		MSG_SET_TMP_FLAGS(msginfo->flags, MSG_INVALID);
 	}
 
-	if (fp) fclose(fp);
+	if (!dest->opened) {
+		procmsg_flush_mark_queue(dest, NULL);
+		procmsg_flush_cache_queue(dest, NULL);
+	}
 
+	S_UNLOCK(mh);
 	return dest->last_num;
 }
 
@@ -612,7 +774,6 @@ static gint mh_move_msg(Folder *folder, FolderItem *dest, MsgInfo *msginfo)
 static gint mh_move_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 {
 	MsgInfo *msginfo;
-	GSList *file_list;
 	gint ret = 0;
 	gint first;
 
@@ -620,12 +781,7 @@ static gint mh_move_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 	if (folder == msginfo->folder->folder)
 		return mh_do_move_msgs(folder, dest, msglist);
 
-	file_list = procmsg_get_message_file_list(msglist);
-	g_return_val_if_fail(file_list != NULL, -1);
-
-	ret = mh_add_msgs(folder, dest, file_list, FALSE, &first);
-
-	procmsg_message_file_list_free(file_list);
+	ret = mh_add_msgs_msginfo(folder, dest, msglist, FALSE, &first);
 
 	if (ret != -1)
 		ret = folder_item_remove_msgs(msginfo->folder, msglist);
@@ -649,7 +805,6 @@ static gint mh_copy_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 {
 	gchar *srcfile;
 	gchar *destfile;
-	FILE *fp;
 	GSList *cur;
 	MsgInfo *msginfo;
 
@@ -661,10 +816,7 @@ static gint mh_copy_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 		if (dest->last_num < 0) return -1;
 	}
 
-	if (dest->opened)
-		fp = NULL;
-	else if ((fp = procmsg_open_mark_file(dest, DATA_APPEND)) == NULL)
-		g_warning(_("Can't open mark file.\n"));
+	S_LOCK(mh);
 
 	for (cur = msglist; cur != NULL; cur = cur->next) {
 		msginfo = (MsgInfo *)cur->data;
@@ -687,7 +839,8 @@ static gint mh_copy_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 			break;
 		}
 
-		g_signal_emit_by_name(syl_app_get(), "add-msg", dest, destfile, dest->last_num + 1);
+		if (syl_app_get())
+			g_signal_emit_by_name(syl_app_get(), "add-msg", dest, destfile, dest->last_num + 1);
 
 		g_free(srcfile);
 		g_free(destfile);
@@ -696,10 +849,8 @@ static gint mh_copy_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 		dest->updated = TRUE;
 		dest->mtime = 0;
 
-		if (fp) {
-			SET_DEST_MSG_FLAGS(fp, dest, dest->last_num,
-					   msginfo->flags);
-		}
+		SET_DEST_MSG_FLAGS(NULL, dest, dest->last_num, msginfo->flags);
+		procmsg_add_cache_queue(dest, dest->last_num, msginfo);
 
 		if (MSG_IS_NEW(msginfo->flags))
 			dest->new++;
@@ -707,8 +858,12 @@ static gint mh_copy_msgs(Folder *folder, FolderItem *dest, GSList *msglist)
 			dest->unread++;
 	}
 
-	if (fp) fclose(fp);
+	if (!dest->opened) {
+		procmsg_flush_mark_queue(dest, NULL);
+		procmsg_flush_cache_queue(dest, NULL);
+	}
 
+	S_UNLOCK(mh);
 	return dest->last_num;
 }
 
@@ -721,11 +876,15 @@ static gint mh_remove_msg(Folder *folder, FolderItem *item, MsgInfo *msginfo)
 	file = mh_fetch_msg(folder, item, msginfo->msgnum);
 	g_return_val_if_fail(file != NULL, -1);
 
-	g_signal_emit_by_name(syl_app_get(), "remove-msg", item, file, msginfo->msgnum);
+	if (syl_app_get())
+		g_signal_emit_by_name(syl_app_get(), "remove-msg", item, file, msginfo->msgnum);
+
+	S_LOCK(mh);
 
 	if (g_unlink(file) < 0) {
 		FILE_OP_ERROR(file, "unlink");
 		g_free(file);
+		S_UNLOCK(mh);
 		return -1;
 	}
 	g_free(file);
@@ -739,6 +898,8 @@ static gint mh_remove_msg(Folder *folder, FolderItem *item, MsgInfo *msginfo)
 		item->unread--;
 	MSG_SET_TMP_FLAGS(msginfo->flags, MSG_INVALID);
 
+	S_UNLOCK(mh);
+
 	if (msginfo->msgnum == item->last_num)
 		mh_scan_folder_full(folder, item, FALSE);
 
@@ -754,7 +915,11 @@ static gint mh_remove_all_msg(Folder *folder, FolderItem *item)
 
 	path = folder_item_get_path(item);
 	g_return_val_if_fail(path != NULL, -1);
-	g_signal_emit_by_name(syl_app_get(), "remove-all-msg", item);
+	if (syl_app_get())
+		g_signal_emit_by_name(syl_app_get(), "remove-all-msg", item);
+
+	S_LOCK(mh);
+
 	val = remove_all_numbered_files(path);
 	g_free(path);
 	if (val == 0) {
@@ -764,6 +929,8 @@ static gint mh_remove_all_msg(Folder *folder, FolderItem *item)
 		item->mtime = 0;
 	}
 
+	S_UNLOCK(mh);
+
 	return val;
 }
 
@@ -771,8 +938,9 @@ static gboolean mh_is_msg_changed(Folder *folder, FolderItem *item,
 				  MsgInfo *msginfo)
 {
 	struct stat s;
+	gchar buf[16];
 
-	if (g_stat(itos(msginfo->msgnum), &s) < 0 ||
+	if (g_stat(utos_buf(buf, msginfo->msgnum), &s) < 0 ||
 	    msginfo->size  != s.st_size ||
 	    msginfo->mtime != s.st_mtime)
 		return TRUE;
@@ -889,10 +1057,16 @@ static gint mh_scan_folder_full(Folder *folder, FolderItem *item,
 
 	debug_print("mh_scan_folder(): Scanning %s ...\n", item->path);
 
+	S_LOCK(mh);
+
 	path = folder_item_get_path(item);
-	g_return_val_if_fail(path != NULL, -1);
+	if (!path) {
+		S_UNLOCK(mh);
+		return -1;
+	}
 	if (change_dir(path) < 0) {
 		g_free(path);
+		S_UNLOCK(mh);
 		return -1;
 	}
 	g_free(path);
@@ -904,6 +1078,7 @@ static gint mh_scan_folder_full(Folder *folder, FolderItem *item,
 	if ((dp = opendir(".")) == NULL) {
 		FILE_OP_ERROR(item->path, "opendir");
 #endif
+		S_UNLOCK(mh);
 		return -1;
 	}
 
@@ -960,6 +1135,10 @@ static gint mh_scan_folder_full(Folder *folder, FolderItem *item,
 		item->new = new;
 		item->unread = unread;
 		item->total = n_msg;
+
+		if (item->cache_queue && !item->opened) {
+			procmsg_flush_cache_queue(item, NULL);
+		}
 	}
 
 	item->updated = TRUE;
@@ -968,6 +1147,7 @@ static gint mh_scan_folder_full(Folder *folder, FolderItem *item,
 	debug_print("Last number in dir %s = %d\n", item->path, max);
 	item->last_num = max;
 
+	S_UNLOCK(mh);
 	return 0;
 }
 
@@ -983,6 +1163,8 @@ static gint mh_scan_tree(Folder *folder)
 
 	g_return_val_if_fail(folder != NULL, -1);
 
+	S_LOCK(mh);
+
 	if (!folder->node) {
 		item = folder_item_new(folder->name, NULL);
 		item->folder = folder;
@@ -993,6 +1175,7 @@ static gint mh_scan_tree(Folder *folder)
 	rootpath = folder_item_get_path(item);
 	if (change_dir(rootpath) < 0) {
 		g_free(rootpath);
+		S_UNLOCK(mh);
 		return -1;
 	}
 	g_free(rootpath);
@@ -1001,6 +1184,7 @@ static gint mh_scan_tree(Folder *folder)
 	mh_remove_missing_folder_items(folder);
 	mh_scan_tree_recursive(item);
 
+	S_UNLOCK(mh);
 	return 0;
 }
 
@@ -1045,6 +1229,7 @@ static gint mh_create_tree(Folder *folder)
 	MAKE_DIR_IF_NOT_EXIST(QUEUE_DIR);
 	MAKE_DIR_IF_NOT_EXIST(DRAFT_DIR);
 	MAKE_DIR_IF_NOT_EXIST(TRASH_DIR);
+	MAKE_DIR_IF_NOT_EXIST(JUNK_DIR);
 
 	return 0;
 }
@@ -1064,6 +1249,8 @@ static FolderItem *mh_create_folder(Folder *folder, FolderItem *parent,
 	g_return_val_if_fail(parent != NULL, NULL);
 	g_return_val_if_fail(name != NULL, NULL);
 
+	S_LOCK(mh);
+
 	path = folder_item_get_path(parent);
 	fs_name = g_filename_from_utf8(name, -1, NULL, NULL, NULL);
 	fullpath = g_strconcat(path, G_DIR_SEPARATOR_S,
@@ -1073,6 +1260,7 @@ static FolderItem *mh_create_folder(Folder *folder, FolderItem *parent,
 
 	if (make_dir_hier(fullpath) < 0) {
 		g_free(fullpath);
+		S_UNLOCK(mh);
 		return NULL;
 	}
 
@@ -1087,6 +1275,7 @@ static FolderItem *mh_create_folder(Folder *folder, FolderItem *parent,
 	folder_item_append(parent, new_item);
 	g_free(path);
 
+	S_UNLOCK(mh);
 	return new_item;
 }
 
@@ -1118,6 +1307,8 @@ static gint mh_move_folder_real(Folder *folder, FolderItem *item,
 		}
 	}
 
+	S_LOCK(mh);
+
 	oldpath = folder_item_get_path(item);
 	if (new_parent) {
 		if (name) {
@@ -1151,6 +1342,7 @@ static gint mh_move_folder_real(Folder *folder, FolderItem *item,
 		g_free(oldpath);
 		g_free(newpath);
 		g_free(utf8_name);
+		S_UNLOCK(mh);
 		return -1;
 	}
 
@@ -1160,6 +1352,7 @@ static gint mh_move_folder_real(Folder *folder, FolderItem *item,
 		g_free(oldpath);
 		g_free(newpath);
 		g_free(utf8_name);
+		S_UNLOCK(mh);
 		return -1;
 	}
 	g_free(rootpath);
@@ -1171,6 +1364,7 @@ static gint mh_move_folder_real(Folder *folder, FolderItem *item,
 		g_free(oldpath);
 		g_free(newpath);
 		g_free(utf8_name);
+		S_UNLOCK(mh);
 		return -1;
 	}
 
@@ -1213,11 +1407,13 @@ static gint mh_move_folder_real(Folder *folder, FolderItem *item,
 	g_free(paths[1]);
 
 	new_id = folder_item_get_identifier(item);
-	g_signal_emit_by_name(syl_app_get(), "move-folder", item, old_id,
-			      new_id);
+	if (syl_app_get())
+		g_signal_emit_by_name(syl_app_get(), "move-folder", item,
+				      old_id, new_id);
 	g_free(new_id);
 	g_free(old_id);
 
+	S_UNLOCK(mh);
 	return 0;
 }
 
@@ -1241,16 +1437,22 @@ static gint mh_remove_folder(Folder *folder, FolderItem *item)
 	g_return_val_if_fail(item != NULL, -1);
 	g_return_val_if_fail(item->path != NULL, -1);
 
+	S_LOCK(mh);
+
 	path = folder_item_get_path(item);
 	if (remove_dir_recursive(path) < 0) {
 		g_warning("can't remove directory `%s'\n", path);
 		g_free(path);
+		S_UNLOCK(mh);
 		return -1;
 	}
 
 	g_free(path);
-	g_signal_emit_by_name(syl_app_get(), "remove-folder", item);
+	if (syl_app_get())
+		g_signal_emit_by_name(syl_app_get(), "remove-folder", item);
 	folder_item_remove(item);
+
+	S_UNLOCK(mh);
 	return 0;
 }
 
@@ -1263,8 +1465,10 @@ static time_t mh_get_mtime(FolderItem *item)
 	path = folder_item_get_path(item);
 	if (g_stat(path, &s) < 0) {
 		FILE_OP_ERROR(path, "stat");
+		g_free(path);
 		return -1;
 	} else {
+		g_free(path);
 		return MAX(s.st_mtime, s.st_ctime);
 	}
 }
@@ -1279,8 +1483,12 @@ static GSList *mh_get_uncached_msgs(GHashTable *msg_table, FolderItem *item)
 	MsgInfo *msginfo;
 	gint n_newmsg = 0;
 	gint num;
+	Folder *folder;
 
 	g_return_val_if_fail(item != NULL, NULL);
+	g_return_val_if_fail(item->folder != NULL, NULL);
+
+	folder = item->folder;
 
 	path = folder_item_get_path(item);
 	g_return_val_if_fail(path != NULL, NULL);
@@ -1298,6 +1506,8 @@ static GSList *mh_get_uncached_msgs(GHashTable *msg_table, FolderItem *item)
 	debug_print("Searching uncached messages...\n");
 
 	if (msg_table) {
+		gint count = 0;
+
 		while ((dir_name = g_dir_read_name(dp)) != NULL) {
 			if ((num = to_number(dir_name)) <= 0) continue;
 
@@ -1320,6 +1530,10 @@ static GSList *mh_get_uncached_msgs(GHashTable *msg_table, FolderItem *item)
 				}
 				n_newmsg++;
 			}
+
+			count++;
+			if (folder->ui_func)
+				folder->ui_func(folder, item, folder->ui_func_data ? folder->ui_func_data : GINT_TO_POINTER(count));
 		}
 	} else {
 		/* discard all previous cache */
@@ -1336,6 +1550,8 @@ static GSList *mh_get_uncached_msgs(GHashTable *msg_table, FolderItem *item)
 				last = last->next;
 			}
 			n_newmsg++;
+			if (folder->ui_func)
+				folder->ui_func(folder, item, folder->ui_func_data ? folder->ui_func_data : GINT_TO_POINTER(n_newmsg));
 		}
 	}
 
@@ -1605,6 +1821,10 @@ static void mh_scan_tree_recursive(FolderItem *item)
 					   !strcmp(dir_name, TRASH_DIR)) {
 					new_item->stype = F_TRASH;
 					folder->trash = new_item;
+				} else if (!folder_get_junk(folder) &&
+					   !strcmp(dir_name, JUNK_DIR)) {
+					new_item->stype = F_JUNK;
+					folder_set_junk(folder, new_item);
 				}
 			}
 
