@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2006 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,6 +31,8 @@
 #include <gtk/gtkmenuitem.h>
 #include <gtk/gtkcheckmenuitem.h>
 #include <gtk/gtkstatusbar.h>
+#include <gtk/gtktooltips.h>
+#include <gtk/gtkarrow.h>
 #include <stdio.h>
 #include <ctype.h>
 #include <string.h>
@@ -63,11 +65,14 @@
 #include "gtkutils.h"
 #include "utils.h"
 #include "rfc2015.h"
+#include "plugin.h"
 
 static GList *messageview_list = NULL;
 
 static void messageview_change_view_type(MessageView		*messageview,
 					 MessageType		 type);
+static void messageview_set_tool_menu	(MessageView		*messageview,
+					 MimeInfo		*mimeinfo);
 static void messageview_set_menu_state	(MessageView		*messageview);
 
 static void messageview_set_encoding_menu
@@ -78,10 +83,18 @@ static gint messageview_delete_cb	(GtkWidget		*widget,
 					 MessageView		*messageview);
 static void messageview_size_allocate_cb(GtkWidget		*widget,
 					 GtkAllocation		*allocation);
+static void messageview_toggle_attach_cb(GtkToggleButton	*toggle,
+					 MessageView		*messageview);
 static void messageview_switch_page_cb	(GtkNotebook		*notebook,
 					 GtkNotebookPage	*page,
 					 guint			 page_num,
 					 MessageView		*messageview);
+static gint messageview_menu_tool_btn_pressed
+					(GtkWidget		*widget,
+					 GdkEventButton		*event,
+					 MessageView		*messageview);
+static void messageview_save_all_cb	(GtkWidget		*widget,
+					 MessageView		*messageview);
 static gboolean key_pressed		(GtkWidget		*widget,
 					 GdkEventKey		*event,
 					 MessageView		*messageview);
@@ -309,9 +322,19 @@ MessageView *messageview_create(void)
 	MessageView *messageview;
 	GtkWidget *vbox;
 	GtkWidget *notebook;
+	GtkWidget *hbox;
+	GtkWidget *toolbar_vbox;
+	GtkWidget *toolbar_hbox;
+	GtkWidget *image;
+	GtkWidget *arrow;
+	GtkWidget *mime_toggle_btn;
+	GtkWidget *menu_tool_btn;
+	GtkWidget *tool_menu;
+	GtkWidget *menuitem;
 	HeaderView *headerview;
 	TextView *textview;
 	MimeView *mimeview;
+	GtkTooltips *tip;
 
 	debug_print(_("Creating message view...\n"));
 	messageview = g_new0(MessageView, 1);
@@ -348,9 +371,59 @@ MessageView *messageview_create(void)
 	gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), 0);
 	gtk_widget_show_all(notebook);
 
-	vbox = gtk_vbox_new(FALSE, 0);
-	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET_PTR(headerview),
+	/* Attachment toolbar */
+	toolbar_vbox = gtk_vbox_new(FALSE, 0);
+	gtk_widget_show(toolbar_vbox);
+
+	toolbar_hbox = gtk_hbox_new(FALSE, 0);
+	gtk_widget_show(toolbar_hbox);
+	gtk_box_pack_end(GTK_BOX(toolbar_vbox), toolbar_hbox, FALSE, FALSE, 0);
+
+	mime_toggle_btn = gtk_toggle_button_new();
+	gtk_widget_show(mime_toggle_btn);
+	gtk_box_pack_start(GTK_BOX(toolbar_hbox), mime_toggle_btn,
 			   FALSE, FALSE, 0);
+	g_signal_connect(G_OBJECT(mime_toggle_btn), "toggled",
+			 G_CALLBACK(messageview_toggle_attach_cb), messageview);
+
+	image = stock_pixbuf_widget_scale(NULL, STOCK_PIXMAP_MAIL_ATTACH, 18, 18);
+	gtk_widget_show(image);
+	gtk_container_add(GTK_CONTAINER(mime_toggle_btn), image);
+
+	tip = gtk_tooltips_new();
+	gtk_tooltips_set_tip(tip, mime_toggle_btn,
+			     _("Switch to attachment list view"), NULL);
+
+	arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
+	gtk_widget_show(arrow);
+	gtk_widget_set_size_request(arrow, 5, -1);
+
+	menu_tool_btn = gtk_button_new();
+	gtk_container_add(GTK_CONTAINER(menu_tool_btn), arrow);
+	GTK_WIDGET_UNSET_FLAGS(menu_tool_btn, GTK_CAN_FOCUS);
+	gtk_widget_show(menu_tool_btn);
+	gtk_box_pack_start(GTK_BOX(toolbar_hbox), menu_tool_btn,
+			   FALSE, FALSE, 0);
+	g_signal_connect(G_OBJECT(menu_tool_btn), "button_press_event",
+			 G_CALLBACK(messageview_menu_tool_btn_pressed),
+			 messageview);
+
+	tool_menu = gtk_menu_new();
+	MENUITEM_ADD(tool_menu, menuitem, NULL, 0);
+	MENUITEM_ADD_WITH_MNEMONIC(tool_menu, menuitem, _("Save _all attachments..."), 0);
+	g_signal_connect(G_OBJECT(menuitem), "activate",
+			 G_CALLBACK(messageview_save_all_cb), messageview);
+	gtk_widget_show_all(tool_menu);
+
+	gtk_widget_hide(toolbar_vbox);
+
+	hbox = gtk_hbox_new(FALSE, 0);
+	gtk_widget_show(hbox);
+	gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET_PTR(headerview), TRUE, TRUE, 0);
+	gtk_box_pack_end(GTK_BOX(hbox), toolbar_vbox, FALSE, FALSE, 0);
+
+	vbox = gtk_vbox_new(FALSE, 0);
+	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
 	gtk_box_pack_start(GTK_BOX(vbox), notebook, TRUE, TRUE, 0);
 	gtk_widget_show(vbox);
 
@@ -362,8 +435,15 @@ MessageView *messageview_create(void)
 			 G_CALLBACK(messageview_switch_page_cb), messageview);
 
 	messageview->vbox        = vbox;
+
 	messageview->notebook    = notebook;
 
+	messageview->hbox            = hbox;
+	messageview->toolbar_vbox    = toolbar_vbox;
+	messageview->mime_toggle_btn = mime_toggle_btn;
+	messageview->menu_tool_btn   = menu_tool_btn;
+	messageview->tool_menu       = tool_menu;
+
 	messageview->new_window  = FALSE;
 	messageview->window      = NULL;
 	messageview->window_vbox = NULL;
@@ -448,6 +528,7 @@ MessageView *messageview_create_with_new_window(void)
 	msgview->visible = TRUE;
 
 	gtk_widget_show_all(window);
+	gtk_widget_hide(msgview->toolbar_vbox);
 
 	messageview_init(msgview);
 
@@ -524,6 +605,7 @@ gint messageview_show(MessageView *messageview, MsgInfo *msginfo,
 	     mimeinfo->mime_type != MIME_TEXT_HTML)) {
 		messageview_change_view_type(messageview, MVIEW_MIME);
 		mimeview_show_message(messageview->mimeview, mimeinfo, file);
+		messageview_set_tool_menu(messageview, mimeinfo);
 	} else {
 		messageview_change_view_type(messageview, MVIEW_TEXT);
 		textview_show_message(messageview->textview, mimeinfo, file);
@@ -533,6 +615,9 @@ gint messageview_show(MessageView *messageview, MsgInfo *msginfo,
 	if (messageview->new_window)
 		messageview_set_menu_state(messageview);
 
+	syl_plugin_signal_emit("messageview-show", messageview, msginfo,
+			       all_headers);
+
 	g_free(file);
 
 	return 0;
@@ -546,12 +631,22 @@ static void messageview_change_view_type(MessageView *messageview,
 	if (messageview->type == type) return;
 
 	if (type == MVIEW_MIME) {
-		gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), TRUE);
+		if (prefs_common.show_attach_tab)
+			gtk_widget_hide(messageview->toolbar_vbox);
+		else
+			gtk_widget_show(messageview->toolbar_vbox);
+		gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook),
+					   prefs_common.show_attach_tab);
 		gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook),
 					      messageview->current_page);
+		if (messageview->current_page == 0)
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(messageview->mime_toggle_btn), FALSE);
+		else
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(messageview->mime_toggle_btn), TRUE);
 	} else if (type == MVIEW_TEXT) {
 		gint current_page = messageview->current_page;
 
+		gtk_widget_hide(messageview->toolbar_vbox);
 		gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
 		gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), 0);
 		messageview->current_page = current_page;
@@ -562,6 +657,45 @@ static void messageview_change_view_type(MessageView *messageview,
 	messageview->type = type;
 }
 
+static void messageview_set_tool_menu(MessageView *messageview,
+				      MimeInfo *mimeinfo)
+{
+	MimeInfo *partinfo;
+	const gchar *name;
+	GtkWidget *menuitem;
+	GList *cur_item, *next = NULL;
+	gint pos = 0;
+	gboolean exist;
+
+	cur_item = GTK_MENU_SHELL(messageview->tool_menu)->children;
+	while (cur_item != NULL) {
+		next = cur_item->next;
+		if (GTK_BIN(cur_item->data)->child == NULL)
+			break;
+		gtk_widget_destroy(GTK_WIDGET(cur_item->data));
+		cur_item = next;
+	}
+
+	for (partinfo = mimeinfo; partinfo != NULL;
+	     partinfo = procmime_mimeinfo_next(partinfo)) {
+		if (partinfo->filename || partinfo->name) {
+			name = g_basename(partinfo->filename ? partinfo->filename : partinfo->name);
+			if (*name == '\0')
+				name = "mimetmp";
+			menuitem = gtk_menu_item_new_with_label(name);
+			gtk_widget_show(menuitem);
+			gtk_widget_set_sensitive(menuitem, FALSE);
+			gtk_menu_shell_insert(GTK_MENU_SHELL(messageview->tool_menu), menuitem, pos++);
+		}
+	}
+
+	exist = (pos != 0);
+	for (cur_item = next; cur_item != NULL; cur_item = cur_item->next) {
+		menuitem = GTK_WIDGET(cur_item->data);
+		gtk_widget_set_sensitive(menuitem, exist);
+	}
+}
+
 static void messageview_set_menu_state(MessageView *messageview)
 {
 	GtkItemFactory *ifactory;
@@ -642,6 +776,7 @@ void messageview_destroy(MessageView *messageview)
 
 	procmsg_msginfo_free(messageview->msginfo);
 
+	gtk_widget_destroy(messageview->tool_menu);
 	if (messageview->window)
 		gtk_widget_destroy(messageview->window);
 
@@ -752,18 +887,27 @@ void messageview_save_as(MessageView *messageview)
 	if (!messageview->msginfo) return;
 	msginfo = messageview->msginfo;
 
-	if (msginfo->subject) {
-		Xstrdup_a(filename, msginfo->subject, return);
-		subst_for_filename(filename);
+	if (msginfo->subject && *msginfo->subject) {
+		filename = g_strdup_printf("%s.eml", msginfo->subject);
+	} else {
+		filename = g_strdup_printf("%u.eml", msginfo->msgnum);
 	}
+	subst_for_filename(filename);
 
 	dest = filesel_save_as(filename);
-	if (!dest) return;
+
+	g_free(filename);
+	if (!dest)
+		return;
 
 	src = procmsg_get_message_file(msginfo);
 	if (copy_file(src, dest, TRUE) < 0) {
+		gchar *utf8_dest;
+
+		utf8_dest = conv_filename_to_utf8(dest);
 		alertpanel_error(_("Can't save the file `%s'."),
-				 g_basename(dest));
+				 g_basename(utf8_dest));
+		g_free(utf8_dest);
 	}
 	g_free(src);
 
@@ -786,11 +930,45 @@ static void messageview_size_allocate_cb(GtkWidget *widget,
 	prefs_common.msgwin_height = allocation->height;
 }
 
+static void messageview_toggle_attach_cb(GtkToggleButton *toggle,
+					 MessageView *messageview)
+{
+	if (gtk_toggle_button_get_active(toggle))
+		gtk_notebook_set_current_page(GTK_NOTEBOOK(messageview->notebook), 1);
+	else
+		gtk_notebook_set_current_page(GTK_NOTEBOOK(messageview->notebook), 0);
+}
+
 static void messageview_switch_page_cb(GtkNotebook *notebook,
 				       GtkNotebookPage *page, guint page_num,
 				       MessageView *messageview)
 {
-	messageview->current_page = page_num;
+	if (messageview->current_page != page_num) {
+		messageview->current_page = page_num;
+		if (page_num == 0)
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(messageview->mime_toggle_btn), FALSE);
+		else
+			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(messageview->mime_toggle_btn), TRUE);
+	}
+}
+
+static gint messageview_menu_tool_btn_pressed(GtkWidget *widget,
+					      GdkEventButton *event,
+					      MessageView *messageview)
+{
+	if (!event)
+		return FALSE;
+
+	gtk_menu_popup(GTK_MENU(messageview->tool_menu), NULL, NULL, 
+		       menu_button_position, messageview->mime_toggle_btn,
+		       event->button, event->time);
+
+	return TRUE;
+}
+
+static void messageview_save_all_cb(GtkWidget *widget, MessageView *messageview)
+{
+	mimeview_save_all(messageview->mimeview);
 }
 
 static gboolean key_pressed(GtkWidget *widget, GdkEventKey *event,
@@ -994,10 +1172,11 @@ static void add_address_cb(gpointer data, guint action, GtkWidget *widget)
 
 	if (!messageview->msginfo) return;
 	msginfo = messageview->msginfo;
-	Xstrdup_a(from, msginfo->from, return);
+	from = g_strdup(msginfo->from);
 	eliminate_address_comment(from);
 	extract_address(from);
 	addressbook_add_contact(msginfo->fromname, from, NULL);
+	g_free(from);
 }
 
 static void create_filter_cb(gpointer data, guint action, GtkWidget *widget)
@@ -1010,7 +1189,7 @@ static void create_filter_cb(gpointer data, guint action, GtkWidget *widget)
 
 	filter_get_keyword_from_msg(messageview->msginfo, &header, &key,
 				    (FilterCreateType)action);
-	prefs_filter_open(messageview->msginfo, header);
+	prefs_filter_open(messageview->msginfo, header, key);
 
 	g_free(header);
 	g_free(key);
