@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2010 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -412,6 +412,7 @@ void mimeview_clear(MimeView *mimeview)
 {
 	procmime_mimeinfo_free_all(mimeview->mimeinfo);
 	mimeview->mimeinfo = NULL;
+	mimeview->has_attach_file = FALSE;
 
 	gtk_tree_store_clear(mimeview->store);
 	textview_clear(mimeview->textview);
@@ -432,6 +433,7 @@ void mimeview_destroy(MimeView *mimeview)
 	textview_destroy(mimeview->textview);
 	imageview_destroy(mimeview->imageview);
 	procmime_mimeinfo_free_all(mimeview->mimeinfo);
+	g_object_unref(mimeview->popupfactory);
 	g_free(mimeview->file);
 	g_free(mimeview->drag_file);
 	g_free(mimeview);
@@ -534,6 +536,8 @@ static gboolean mimeview_append_part(MimeView *mimeview, MimeInfo *partinfo,
 	mime_type = partinfo->content_type ? partinfo->content_type : "";
 	size = to_human_readable(partinfo->content_size);
 	name = get_part_name(partinfo);
+	if (name && *name != '\0')
+		mimeview->has_attach_file = TRUE;
 
 	gtk_tree_store_append(mimeview->store, iter, parent);
 	gtk_tree_store_set(mimeview->store, iter,
@@ -880,6 +884,9 @@ static gint mimeview_button_pressed(GtkWidget *widget, GdkEventButton *event,
 			menu_set_sensitive(mimeview->popupfactory,
 					   "/Open", TRUE);
 
+		menu_set_sensitive(mimeview->popupfactory,
+				   "/Save all...", mimeview->has_attach_file);
+
 		if (partinfo && (partinfo->mime_type == MIME_TEXT ||
 				 partinfo->mime_type == MIME_TEXT_HTML ||
 				 partinfo->mime_type == MIME_MESSAGE_RFC822))
@@ -1061,8 +1068,6 @@ static void mimeview_display_as_text(MimeView *mimeview)
 
 void mimeview_save_as(MimeView *mimeview)
 {
-	gchar *filename;
-	gchar *defname = NULL;
 	MimeInfo *partinfo;
 
 	if (!mimeview->opened) return;
@@ -1071,21 +1076,7 @@ void mimeview_save_as(MimeView *mimeview)
 	partinfo = mimeview_get_selected_part(mimeview);
 	g_return_if_fail(partinfo != NULL);
 
-	if (partinfo->filename)
-		defname = partinfo->filename;
-	else if (partinfo->name) {
-		Xstrdup_a(defname, partinfo->name, return);
-		subst_for_filename(defname);
-	}
-
-	filename = filesel_save_as(defname);
-	if (!filename) return;
-
-	if (procmime_get_part(filename, mimeview->file, partinfo) < 0)
-		alertpanel_error
-			(_("Can't save the part of multipart message."));
-
-	g_free(filename);
+	mimeview_save_part_as(mimeview, partinfo);
 }
 
 void mimeview_save_all(MimeView *mimeview)
@@ -1111,6 +1102,15 @@ void mimeview_print(MimeView *mimeview)
 	partinfo = mimeview_get_selected_part(mimeview);
 	g_return_if_fail(partinfo != NULL);
 
+	mimeview_print_part(mimeview, partinfo);
+}
+
+void mimeview_print_part(MimeView *mimeview, MimeInfo *partinfo)
+{
+	g_return_if_fail(partinfo != NULL);
+
+	if (!mimeview->file) return;
+
 	if (partinfo->mime_type == MIME_MESSAGE_RFC822) {
 		gchar *filename;
 		MsgInfo *msginfo;
@@ -1137,17 +1137,14 @@ void mimeview_print(MimeView *mimeview)
 	}
 }
 
-static void mimeview_launch(MimeView *mimeview)
+void mimeview_launch_part(MimeView *mimeview, MimeInfo *partinfo)
 {
-	MimeInfo *partinfo;
 	gchar *filename;
 
-	if (!mimeview->opened) return;
-	if (!mimeview->file) return;
-
-	partinfo = mimeview_get_selected_part(mimeview);
 	g_return_if_fail(partinfo != NULL);
 
+	if (!mimeview->file) return;
+
 	filename = procmime_get_tmp_file_name(partinfo);
 
 	if (procmime_get_part(filename, mimeview->file, partinfo) < 0)
@@ -1159,18 +1156,15 @@ static void mimeview_launch(MimeView *mimeview)
 	g_free(filename);
 }
 
-static void mimeview_open_with(MimeView *mimeview)
+void mimeview_open_part_with(MimeView *mimeview, MimeInfo *partinfo)
 {
-	MimeInfo *partinfo;
 	gchar *filename;
 	gchar *cmd;
 
-	if (!mimeview->opened) return;
-	if (!mimeview->file) return;
-
-	partinfo = mimeview_get_selected_part(mimeview);
 	g_return_if_fail(partinfo != NULL);
 
+	if (!mimeview->file) return;
+
 	filename = procmime_get_tmp_file_name(partinfo);
 
 	if (procmime_get_part(filename, mimeview->file, partinfo) < 0) {
@@ -1202,6 +1196,61 @@ static void mimeview_open_with(MimeView *mimeview)
 	g_free(filename);
 }
 
+void mimeview_save_part_as(MimeView *mimeview, MimeInfo *partinfo)
+{
+	gchar *filename = NULL;
+
+	g_return_if_fail(partinfo != NULL);
+
+	if (!mimeview->file) return;
+
+	if (partinfo->filename) {
+		filename = filesel_save_as(partinfo->filename);
+	} else if (partinfo->name) {
+		gchar *defname;
+
+		defname = g_strdup(partinfo->name);
+		subst_for_filename(defname);
+		filename = filesel_save_as(defname);
+		g_free(defname);
+	} else
+		filename = filesel_save_as(NULL);
+
+	if (!filename) return;
+
+	if (procmime_get_part(filename, mimeview->file, partinfo) < 0)
+		alertpanel_error
+			(_("Can't save the part of multipart message."));
+
+	g_free(filename);
+}
+
+static void mimeview_launch(MimeView *mimeview)
+{
+	MimeInfo *partinfo;
+
+	if (!mimeview->opened) return;
+	if (!mimeview->file) return;
+
+	partinfo = mimeview_get_selected_part(mimeview);
+	g_return_if_fail(partinfo != NULL);
+
+	mimeview_launch_part(mimeview, partinfo);
+}
+
+static void mimeview_open_with(MimeView *mimeview)
+{
+	MimeInfo *partinfo;
+
+	if (!mimeview->opened) return;
+	if (!mimeview->file) return;
+
+	partinfo = mimeview_get_selected_part(mimeview);
+	g_return_if_fail(partinfo != NULL);
+
+	mimeview_open_part_with(mimeview, partinfo);
+}
+
 static void mimeview_view_file(const gchar *filename, MimeInfo *partinfo,
 			       const gchar *cmdline)
 {
@@ -1226,6 +1275,21 @@ static void mimeview_view_file(const gchar *filename, MimeInfo *partinfo,
 		}
 		execute_open_file(filename, partinfo->content_type);
 		return;
+#elif defined(__APPLE__)
+		if (g_file_test(filename, G_FILE_TEST_IS_EXECUTABLE) ||
+		    str_has_suffix_case(filename, ".py") ||
+		    str_has_suffix_case(filename, ".rb") ||
+		    str_has_suffix_case(filename, ".sh")) {
+			alertpanel_full
+				(_("Opening executable file"),
+				 _("This is an executable file. Opening executable file is restricted for security.\n"
+				   "If you want to launch it, save it to somewhere and make sure it is not an virus or something like a malicious program."),
+				 ALERT_WARNING, G_ALERTDEFAULT, FALSE,
+				 GTK_STOCK_OK, NULL, NULL);
+			return;
+		}
+		execute_open_file(filename, partinfo->content_type);
+		return;
 #else
 		if (MIME_IMAGE == partinfo->mime_type)
 			cmd = prefs_common.mime_image_viewer;
