@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2006 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -71,6 +71,8 @@ struct _MimeView
 	gchar *file;
 
 	gchar *drag_file;
+
+	gboolean has_attach_file;
 };
 
 MimeView *mimeview_create	(void);
@@ -94,4 +96,14 @@ void mimeview_save_all			(MimeView	*mimeview);
 
 void mimeview_print			(MimeView	*mimeview);
 
+void mimeview_launch_part		(MimeView	*mimeview,
+					 MimeInfo	*partinfo);
+void mimeview_open_part_with		(MimeView	*mimeview,
+					 MimeInfo	*partinfo);
+void mimeview_save_part_as		(MimeView	*mimeview,
+					 MimeInfo	*partinfo);
+
+void mimeview_print_part		(MimeView	*mimeview,
+					 MimeInfo	*partinfo);
+
 #endif /* __MIMEVIEW_H__ */
